package controller;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import controller.model.NewTurn;
import controller.model.NewTurnDetector;
import controller.model.TickInfo;
import controller.model.TurnDetector;
import champ2010client.Action;
import champ2010client.Controller;
import champ2010client.SensorModel;

/**
 * Kontroler vozidla zatacajuci podla najdlhsieho senzora.
 */
public class PrototypeFirstTime extends Controller
{	
	
	//instancia recovery modul
	RecoveryModule recMod;
	//radenie
	//zoradeny rad otacok za poslenych 30 tikov
	private Deque<Double> rpm30Tick = new ArrayDeque<Double>(); 
	private double clutchCur=0;
	boolean RPMGrows=false;
	boolean notOnTrack =false;
	int dontGear =0;// 7 tikov trva kym otacky zacnu po radeni stupat
	/*
	 * Koeficient frikcie medzi pneumatikou a vozovkou; tato hodnota 
	 * sa v prvom kole na zacitaku zistuje pri rychlosti 150 z brzdnej 
	 * drahy a flagy potrebne pri jej urceni
	 */
	private double friction=16;
	private boolean kolesaStoja = false,brzdi =false,vypocitajFriction=false;
	private double brakeStartDist;
	//rychlost pri akej sa ma merat frikcia
	private double breakSpeed=125;
	
	// rad sluziaci na komunikaciu s vlaknom vytvarajucim model
	BlockingQueue<TickInfo> tickInfoPipe;
	
	// rad sluziaci na komunikaciu s vlaknom vytvarajucim model
//	private final BlockingQueue<SensorInfo> sensorInfoPipe;

	//premenna pre sirku trate
	private double trackWidth=0;
	private boolean trackWidthMeasuring = true;
	
	//premenna pre dlzku trate
	private double trackLength=0;
	private boolean trackLengthMeasuring = true;
	
	//premenna pre zaznamenavanie radiusov jednotlivych metrov trate
	private int trackRadius = 1;
	
	private Deque<Integer> turnDirection = new ArrayDeque<Integer>(); 
	
	// potrebne k recovery
	private int ticCount=0;
	private boolean afterStart=false;
	//line assist
	private NewTurn nextTurn = null;
	private Iterator<NewTurn> iterator;
	private LinkedList<NewTurn> turnList;
	private double targetPos=0.4;
	private double lastTurnEnd;

	// premenne potrebne pre zameranie focusu
	private static Boolean requestedChange = false;
	private static int desiredFocus;
	private static double lastChange;
	private static double focusValue = -1;
	public static Object lock = new Object();
	private boolean isFirstLap=true;
	boolean rovinaNemeratZakruty=false;
	
	public PrototypeFirstTime()
	{
		tickInfoPipe = new LinkedBlockingQueue<TickInfo>(1000);
		
		Thread t = new Thread(new NewTurnDetector(tickInfoPipe));
		t.setDaemon(true);
		t.setPriority(Thread.MIN_PRIORITY);
		t.start();

		recMod = new RecoveryModule();

	}
	
	public Action control(SensorModel sm)
	{
		// sleduje sa, ci je uz po starte (100 tickov)
		ticCount++;
		if (ticCount == 100)
			afterStart=true;
		Action action = new Action ();

		checkFocus(sm, action);
		
		double radius;
		//zistenie radiusu
		if ((sm.getDistanceFromStartLine() > trackRadius) && (sm.getDistanceFromStartLine() < (trackRadius+3)))
		{
			trackRadius += 3;
			if (sm.getTrackPosition() > 0)
				radius = computeRadius(sm, 13);
			else
				radius = computeRadius(sm, 3);
			
			if (radius < 2000)
				//System.out.println(radius);
				;
		}
		
	//zistenie sirky trate
		if(isFirstLap)
			isFirstLap=(sm.getLastLapTime()==0.0);
		if (trackWidthMeasuring)
		{
			trackWidth=sm.getTrackEdgeSensors()[0]+sm.getTrackEdgeSensors()[18]; 
			//System.out.println(trackWidth+"sirka trate");
			//System.out.println(sm.getTrackPosition());
			trackWidthMeasuring = false;
		}
		
		//zistenie dlzky trate
		if (trackLengthMeasuring)
		{
			if ((sm.getDistanceFromStartLine() < trackLength) && (trackLength > 0))
			{
				trackLengthMeasuring = false;
				//System.out.println(trackLength+"dlzka trate");
			}
			else
				trackLength=sm.getDistanceFromStartLine();		
		}
		
		
		
		
		/* * * * * R E C O V E R Y * * * Z A C I A T O K * * */

		/* 
		 * v kazdom ticku sa overi, ci sa pilot nenachadza v specifuckych situaciach recovery 
		 * 
		 */
		
		// ak je po starte, kontorluje sa specificka situacia
		if (afterStart)
			recMod.checkRecovery(sm);
		
		// zablokovany pri okraji trate (kolmo) - prechod do noveho recovery stavu
		if ((recMod.isBlocked()) && afterStart && (-1<sm.getTrackPosition() && sm.getTrackPosition()<1)){
			if (sm.getTrackPosition()<0)
				return recMod.recoverRight(sm); 
			else if (sm.getTrackPosition()>0)
				return recMod.recoverLeft(sm);
		}
			
		// zablokovany v kopci 			
		if ((recMod.isBlocked()) && afterStart)
			return recMod.goStraight(sm);

		// vyletel som na lavu stranu cesty alebo bol v recovrery state left a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if ((sm.getTrackPosition()>1 || recMod.isInRecoveryStateLeft()) && (!recMod.isBlocked())){
			notOnTrack=true;
			return recMod.recoverLeft(sm);
		} //else if (recMod.isBlocked())
		
		// vyletel som na pravu stranu cesty alebo bol v recovrery state right a neje v pozicii kedy moze odovzdat auto modulu riadenia
		if ((sm.getTrackPosition()<-1 || recMod.isInRecoveryStateRight()) && (!recMod.isBlocked())){
			notOnTrack=true;
			return recMod.recoverRight(sm);
		} //else if (recMod.isBlocked())

		// som na trati v protismere - natoceny viac do lava 
		if ((sm.getAngleToTrackAxis()>Math.PI/2.0) && (sm.getAngleToTrackAxis()<Math.PI)){
//			System.out.println("som na trati v protismere - natoceny viac do lava");
			return recMod.recoverRight(sm);
		}
		
		// som na trati v protismere - natoceny viac do prava
		if ((sm.getAngleToTrackAxis()<-Math.PI/2.0) && (sm.getAngleToTrackAxis()>-Math.PI)){
//			System.out.println("som na trati v protismere - natoceny viac do prava");
			return recMod.recoverLeft(sm);
		}
		
		/* * * * * R E C O V E R Y * * * K O N I E C * * */		
		
		
		/*
		 * rychlost - akceleracia naplno do dosiahnutia maximalnej rychlosti, ktoru je mozne ubrzdit na vzdialenost
		 * najdlhsieho senzora,
		 * vzorec			v = odmocnina(2 * g * mi * brzdna draha)
		 */
		double speedAdjustment;
		if(isFirstLap)
			speedAdjustment=60;
		else
			speedAdjustment=30;
		
		if (sm.getSpeed () +speedAdjustment  < Math.sqrt(2 * 9.81 * friction * (Math.abs(sm.getTrackEdgeSensors()[9]-trackWidth/2))))
		{	action.accelerate = ESP.filterASR(sm, ESP.actionAccelAktual);
			ESP.actionBrakeAktual = 0;
			ESP.actionBrakeTickBack = 0;
		}else if(sm.getSpeed()<7){
			action.accelerate = 1;
			ESP.actionBrakeAktual = 0;
			ESP.actionBrakeTickBack = 0;
		}
		else{
			action.brake = ESP.filterABS(sm, ESP.actionBrakeAktual);
			ESP.actionAccelAktual = 0;
			ESP.actionAccelTickBack = 0;
		}
	
		// * riadenie *
		// zmena smeru na zaklade najdlhsieho senzora
		//action.steering = steeringAdjustment(computeAvgLongestSensorAngle(sm.getTrackEdgeSensors()));
		action.steering=firstLapSteer(sm);
		
		
//LINE ASSIST
		//*************************************/
		//vyber zakruty zo zoznamu
		//*************************************/
		
		if(sm.getLastLapTime() != 0 && nextTurn==null) { //inicializacia zoznamu zakrut
			turnList=NewTurnDetector.getTurnList();
			if(turnList.get(0).getEndDistance()>turnList.get(1).getStartDistance())
				turnList.remove();
			iterator=turnList.iterator();
			if(iterator.hasNext())						//inicializacia prvej zakruty
				lastTurnEnd=0;
				nextTurn=iterator.next();
				
		}
		if(sm.getLastLapTime() != 0 && !iterator.hasNext()) //obnovenie iteratora zoznamu zakrut
			iterator=turnList.iterator();
		
		if(sm.getLastLapTime() != 0 && nextTurn != null && (sm.getDistanceFromStartLine()-nextTurn.getEndDistance())>0 && (sm.getDistanceFromStartLine()-nextTurn.getEndDistance())<5) {
			lastTurnEnd=nextTurn.getEndDistance();
			nextTurn=iterator.next();
		}
		
		//*************************************/
		//korekcia steeringu podla typu zakruty - prudke zakruty
		//*************************************/
		
		if(sm.getLastLapTime() != 0 && (nextTurn.getStartDistance()-sm.getDistanceFromStartLine())>5 && (nextTurn.getStartDistance()-sm.getDistanceFromStartLine())<200 && nextTurn.getPeakValue()>0.25) {
					if((nextTurn.getStartDistance()-lastTurnEnd)>50)
						if(nextTurn.getDirection() == controller.model.NewTurn.Direction.RIGHT) { //line assist pred zakrutou vpravo
							System.out.println("LINE ASSIST RIGHT");	
							action=navigateToPosition(action,sm, nextTurn.getStartDistance(),targetPos);
							if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
								action.steering=0;
							if(action.steering>0.2)
								action.steering+=0.2;
							if(action.steering<-0.2)
								action.steering+=-0.2;
						}
						else { //line assist pred zakrutou vlavo
							System.out.println("LINE ASSIST LEFT"); 
							action=navigateToPosition(action,sm, nextTurn.getStartDistance(),-targetPos);
							if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
								action.steering=0;
							if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
								action.steering=0;
							if(action.steering>0.2)
								action.steering+=0.2;
							if(action.steering<-0.2)
								action.steering+=-0.2;
						}
		}
		
		//*************************************/
		//korekcia steeringu podla typu zakruty - mierne zakruty - TODO
		//*************************************/
		
		/*if(sm.getLastLapTime() != 0 && (nextTurn.getStartDistance()-sm.getDistanceFromStartLine())>-10 && (nextTurn.getStartDistance()-sm.getDistanceFromStartLine())<200 && nextTurn.getPeakValue()<0.2) {
			if((nextTurn.getStartDistance()-lastTurnEnd)>20)
				if(nextTurn.getDirection() == controller.model.NewTurn.Direction.RIGHT) { //line assist pred zakrutou vpravo
					System.out.println("LINE ASSIST RIGHT");	
					action=navigateToPosition(action,sm, nextTurn.getStartDistance(),-targetPos);
					if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
						action.steering=0;
					if(action.steering>0.2)
						action.steering+=0.2;
					if(action.steering<-0.2)
						action.steering+=-0.2;
					
				}
				else { //line assist pred zakrutou vlavo
					System.out.println("LINE ASSIST LEFT"); 
					action=navigateToPosition(action,sm, nextTurn.getStartDistance(),targetPos);
					if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
						action.steering=0;
					if((action.steering>0 && action.steering<0.015) || (action.steering<0 && action.steering>-0.015))
						action.steering=0;
					if(action.steering>0.2)
						action.steering+=0.2;
					if(action.steering<-0.2)
						action.steering+=-0.2;
				}
		}*/
		
		//LINE ASSIST
		//*************************************/
		//KONIEC
		//*************************************/
		
		if(sm.getLastLapTime()>0 && false){
			if (sm.getSpeed () -60  < countSpeed(action.steering))
			{	action.accelerate = ESP.filterASR(sm, ESP.actionAccelAktual);
				ESP.actionBrakeAktual = 0;
				ESP.actionBrakeTickBack = 0;
			}else if(sm.getSpeed()<7){
				action.accelerate = 1;
				ESP.actionBrakeAktual = 0;
				ESP.actionBrakeTickBack = 0;
			}
			else{
				action.brake = ESP.filterABS(sm, ESP.actionBrakeAktual);
				ESP.actionAccelAktual = 0;
				ESP.actionAccelTickBack = 0;
			}
		}
		action = getGear(sm,action,rpm30Tick);

		
		try {
			tickInfoPipe.put(new TickInfo(action, sm, !rovinaNemeratZakruty));
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri posielani akcie do modelu trate.");
			e.printStackTrace();
		}


		if(friction ==16 ){
			friction=getFriction(sm,action);
			ESP.actionAccelAktual = 1;
			ESP.actionAccelTickBack = 1;
		}
	/*if(sm.getSpeed()>100)
			System.out.println(sm.getCurrentLapTime());*/
		//System.out.println(Math.toDegrees(sm.getAngleToTrackAxis()));
		return action;
	}
	
	/**
	 * Metoda urcena na vyziadanie si zmeny focusu.
	 * @param desiredFocus zelane nastavenie focusu z rozsahu <-90, 90>
	 */
	public static void requestFocusChange(int desiredFocus) {
		synchronized (requestedChange)
		{
			PrototypeFirstTime.desiredFocus = desiredFocus;
			requestedChange = true;
		}
	}
	/**
	 * metoda na nastavanie steeringu v prvom kole, zatial sa vyuziva ja v ostatnych kolach,
	 * okrem to ho ze vracia steering upravuje a hodnoty globalnych premenny ako rovinaNemeratZakruty, turnDirection 	
	 *	
	 * @param sm senzor model
	 * @return vracia steerin 
	 */
	private double firstLapSteer(SensorModel sm){
		double steer=0;
		int max=0;
		for (int i = 1; i < sm.getTrackEdgeSensors().length; i++)
			if (sm.getTrackEdgeSensors()[max] < sm.getTrackEdgeSensors()[i])
				max = i;
		int uhol= 0;
		double whenToSteer;
		if(isFirstLap)
			whenToSteer=3*trackWidth;
		else
			whenToSteer=200;	
		//boolean do optimalizacie 
		rovinaNemeratZakruty=false;
		
		//System.out.println("lavo: "+sm.getTrackEdgeSensors()[8]+"stredny: "+sm.getTrackEdgeSensors()[9]+" pravo: "+sm.getTrackEdgeSensors()[10]);
		/*if(sm.getTrackEdgeSensors()[9]<13+trackWidth && isFirstLap)
			uhol=turnDirection.getFirst();
		else*/
		if(max<9  && ((sm.getTrackEdgeSensors()[max]-sm.getTrackEdgeSensors()[9])>trackWidth || sm.getTrackEdgeSensors()[max]<2*trackWidth || turnDirection.getFirst()!=0) && sm.getTrackEdgeSensors()[9]<whenToSteer){
			uhol=1;
			//if(sm.getTrackPosition()<0)
			//	lol=0.95+sm.getTrackPosition();
			//System.out.println("stredny:"+sm.getTrackEdgeSensors()[9]+"bocny");
		}
		else if(max>9 && ((sm.getTrackEdgeSensors()[max]-sm.getTrackEdgeSensors()[9])>trackWidth || sm.getTrackEdgeSensors()[max]<2*trackWidth || turnDirection.getFirst()!=0) && sm.getTrackEdgeSensors()[9]<whenToSteer){
			uhol=-1;
			//if(sm.getTrackPosition()>0)
			//	lol=0.95-sm.getTrackPosition();
		}
		else{
			uhol =0;
		}
		
		if(turnDirection.size()==3)
			turnDirection.pollFirst();
		turnDirection.addLast(uhol);
		uhol =getTurnDirection(turnDirection);
			if(uhol==0)
				rovinaNemeratZakruty=true;
		if(!(sm.getAngleToTrackAxis()>-0.001 && sm.getAngleToTrackAxis()<0.001)&& uhol !=0 ){
			steer=(steeringAdjustmentNew(sm.getTrackEdgeSensors()[9],(trackWidth),0)*(uhol) -(sm.getTrackPosition()*0.1)/0.366519);
		}else {
			//System.out.println("rovno");
			//aby sel na stred doplnit -sm.getTrackPosition()*0.1
			
			/*if(sm.getTrackEdgeSensors()[9]>60)
				steer=(sm.getAngleToTrackAxis()-sm.getTrackPosition()*0.1)/0.366519;
			else if(sm.getTrackEdgeSensors()[9]<(13+trackWidth))
				steer=0.005*uhol;*/
			
			if(!isFirstLap && sm.getTrackEdgeSensors()[max]<2*trackWidth)
				steer=(sm.getAngleToTrackAxis())/0.366519;
			else
				steer=(sm.getAngleToTrackAxis()-sm.getTrackPosition()*0.1)/0.366519;
		}
		
	/*	if(sm.getTrackPosition()>0.75 && uhol == -1 && !isFirstLap)
			steer*=1.2;
		else if(sm.getTrackPosition()<-0.75 && uhol == 1 && !isFirstLap)
			steer*=1.2;*/
		return steer;
	}
	/**
	 * Metoda sluzi na ziskanie novej hodnoty focusu po zavolani metody
	 * requestFocusChange. Metoda ma byt zavolana z metody control.
	 * @param sm senzor model
	 * @param action vytvarana akcia
	 */
	private void checkFocus(SensorModel sm, Action action) {
		// prestavenie udaju o poslednej zmene pri prejazde startom
		if (sm.getCurrentLapTime() < lastChange)
			lastChange = sm.getCurrentLapTime() - sm.getLastLapTime();
		// nacitanie novej hodnoty focusu a oznamenie cakajucim jej ziskanie
		if (focusValue == -2)
		{
			focusValue = sm.getFocusSensors()[2];
			synchronized (lock)
			{
				lock.notify();
				TurnDetector.setFocusValue(focusValue);
			}
		}
		//zmena focusu
		synchronized (requestedChange)
		{
			if (requestedChange)
				if (sm.getCurrentLapTime() > lastChange + 1) // focus je mozne menit kazdu 1s
				{
					action.focus = desiredFocus;
					requestedChange = false;
					lastChange = sm.getCurrentLapTime();
					focusValue = -2; // -2 vo focusValue znamena, ze chcem novu hodnotu
				}
				else
				{
					requestedChange = false;
					TurnDetector.setFocusValue(-3); // -3 znamena priskore volanie
					synchronized (lock)
					{
						lock.notify();
					}
				}
		}
	}
	
	/**
	 * metoda na zistenie frikcie z brzdnej drahy pri zaseknutych kolesach(vsetky4),
	 *  pouziva v sebe globalne premenne breakSpeed(150), breakStartDist a tri 
	 *  flagy a to vypocitajFriction, brzdi, kolesaStoja
	 * 
	 * @param sensors aktualne senzory
	 * @param action aktualna akcia
	 * @return 0 ak auto este nezastavilo a frikciu pre povrch ak auto zastalo
	 */
    private double getFriction(SensorModel sensors, Action action){
		
		
		if(sensors.getSpeed()<=breakSpeed && !brzdi){
			action.accelerate=1;
			action.brake=0;
			brakeStartDist = sensors.getDistanceRaced();
			//System.out.println(brakeS);
		}else{
			action.steering=0;
			double speed =0;
			//System.out.print(zrychlenie(sensorsBefore.getSpeed(), sensors.getSpeed(), sensorsBefore.getCurrentLapTime(), sensors.getCurrentLapTime()));
			//System.out.print(" =||= ");
			for(int i=0;i<4;i++){
				speed+=sensors.getWheelSpinVelocity()[i]*0.3179;
				//System.out.print(" || "+zrychlenie(sensorsBefore.getWheelSpinVelocity()[i]*0.3179, sensors.getWheelSpinVelocity()[i]*0.3179, sensorsBefore.getCurrentLapTime(), sensors.getCurrentLapTime()));
			}
			speed/=4;
			//System.out.println();
			//System.out.println("rychlost kolies: "+speed +" brakespeed "+breakSpeed+" brakedist "+brakeStartDist);
			
			if(speed<0.05 && !kolesaStoja){
				brakeStartDist= sensors.getDistanceRaced();
				kolesaStoja=true;
				breakSpeed=sensors.getSpeed();
			}
			brzdi=true;
			action.accelerate=0;
			action.brake=1;
			action.clutch=0;
			if(sensors.getSpeed()<0.01 && !vypocitajFriction){
				//System.out.println("drzdna draha"+(sensors.getDistanceRaced()-brakeStartDist)+ " rychlost kolies: "+speed +" brakespeed "+breakSpeed+" brakedist "+brakeStartDist+" friction "+(Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist))));
				vypocitajFriction=true;
				System.out.println(Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist)));
				return (Math.pow(breakSpeed, 2)/(2*9.81*(sensors.getDistanceRaced()-brakeStartDist)));
				
			}//System.out.println(sensors.getSpeed());
			
		}
		return 16;
	}

	/**
	 * metoda urcujuca polomer useku trate na zaklade troch trackEdge senyorov
	 * 
	 * @param sensors senzormodel
	 * @param first prvy z trojice senzorov, od ktoreho sa pocita radius
	 * @return radius useku trate
	 */
	private double computeRadius(SensorModel sensors, int first)
	{
		double x1 = Math.cos(Math.PI/6) * sensors.getTrackEdgeSensors()[first];
		double y1 = Math.sin(Math.PI/6) * sensors.getTrackEdgeSensors()[first];
		double x2 = Math.cos(Math.PI/4.5) * sensors.getTrackEdgeSensors()[first+1];
		double y2 = Math.sin(Math.PI/4.5) * sensors.getTrackEdgeSensors()[first+1];
		double x3 = Math.cos(Math.PI/3.6) * sensors.getTrackEdgeSensors()[first+2];
		double y3 = Math.sin(Math.PI/3.6) * sensors.getTrackEdgeSensors()[first+2];
		
		double a,b,c,s;
		
		a = Math.sqrt(Math.pow((x1-x2), 2)+Math.pow((y1-y2), 2));
		b = Math.sqrt(Math.pow((x2-x3), 2)+Math.pow((y2-y3), 2));
		c = Math.sqrt(Math.pow((x3-x1), 2)+Math.pow((y3-y1), 2));
		s = (a+b+c)/2;
		
		return (a*b*c)/(4*Math.sqrt(s*(s-a)*(s-b)*(s-c)));
	}
	
	/**
	 * metoda modulu riadenia na urcenie prevodoveho stupna s otacok motora a podla toho ci rastu/klesaju
	 * 
	 * @param sensors senzormodel
	 * @param action akcia na vykonanie
	 * @param rpmGrows boolena hovorici o tom ci otacky klesaju alebo stupaju
	 * @return akciu, ktora ma nastaveny aj prevodovy stupen 
	 */
	private Action getGear(SensorModel sensors,Action action, Deque<Double> rad){
		int gearCur;
		double rpm = sensors.getRPM(),lastRpm=9000;
		int[]  gearUp = {9500, 9500, 9500, 9500, 9500, 0};
		int[]  gearDown =  {0, 4300, 6200, 7000, 7300, 7700};
		gearCur=sensors.getGear();//berem akt stupen
		
		//updatne sa zoznam otacok za poslednych 30 tikov
		if(rad.size()>0)
			lastRpm=rad.getLast();
		if(rad.size()==30 && dontGear==0)
			rad.pollFirst();
		if(dontGear==0)
			rad.addLast(rpm);
		else
			dontGear--;
		
		//som na starte a pretek este nebol odstarovany radim 1, stlacim spojku, pridam plyn 
		if(gearCur<1){
			gearCur=1;
			action.gear=gearCur;
			action.accelerate= 1;
			clutchCur=0.9;
			//dontGear=8;
		}
		//uz je odstartovane a mam stlacenu spojku, zaradenu 1, ak otack rastu a su nad  9000 tak uberam spojku po 0.05, ak klesnu pod 7000 uberam iba po 0.02 ak klesnu pod 6500 pridavam k spojke po 0.01
		//podmienka clutchCur>0.67 a akcia clutchCur-=0.34;
		//TODO ked bude asr s nastavitelni smykom tak zmeni podmienku s casom a dat namiesto acc 1 asr
		else if(((clutchCur>0.05 && gearCur==1 )|| sensors.getSpeed()<55) && sensors.getCurrentLapTime()<3.5 && sensors.getLastLapTime()==0){
			action.accelerate=1;
			//if((rpm-lastRpm)>=0 && rpm > 9300 && sensors.getSpeed()<40)
			//	clutchCur-=0.08;
			//else 
			if((rpm-lastRpm)>=0 && rpm > 9000)
				clutchCur-=0.05;
			else if((rpm-lastRpm)>=0 && rpm > 7000)
				clutchCur-=0.02;
			else if(rpm < 6500)
				clutchCur+=0.01;
			//System.out.println("som tu");
		}else if(clutchCur>0 && (rpm-lastRpm)>0 && rpm>7000 && gearCur==1 && rpm<9499){
			clutchCur-=0.03;
			//System.out.println("lol");
		}
		// rychlost je mensia ako 6 sucastne su otaky vecsie/rovne otackam ked radim, a sucasne otacky v predchadzajucich 30 tikoch stupali => radim hore
		else if (gearCur < 6 && rpm >= gearUp[gearCur-1] && dontGear==0)
		{	
			
			if(gearCur==1)
				clutchCur=0.63;
			if(isDequeGrowing(rad)){
				dontGear=8;
				gearCur++;
			}
			
			
		}
		// rychlost je vetsia ako 1 sucastne su otaky mensie/rovne otackam ked podradzujem a otacky za poslednych 30 tikov klesaju
		else if (gearCur > 1 && rpm <= gearDown[gearCur-1]  && dontGear==0 )
		{
			if(!isDequeGrowing(rad)){
				dontGear=8;
				gearCur--;	
			}
			
		}
			
		//mam zaradenu inu rychlost ako 1 a spojka je stlacena, ak je zaradena dvojka tak pustam spojku ak su otacky nad 6500 ak je ina rychlost tak pustam spojku
		if((clutchCur>0 && gearCur>1 && (rpm-lastRpm)>0) || rpm>9000 && gearCur>1){
			if(gearCur==2 && rpm>7000)
				clutchCur-=0.01;
			else if(gearCur==2 && rpm>8500)
				clutchCur-=0.08;
			else if (gearCur!=2){
				clutchCur-=0.15;
				
			}
		}
		//System.out.println(rpm);
		if(sensors.getCurrentLapTime()<6 && sensors.getLastLapTime()==0)
			action.accelerate=1;
		action.gear=gearCur;
		action.clutch=clutchCur;
		return action;
	}
	/**
	 * metoda vrati true ak rad<double> apon s polovice rastie a false ak nsatane opacne situacia
	 * 
	 * @param que rad cisel ktorych rast sa ma urcit
	 * @return true alebo false
	 */
	private boolean isDequeGrowing(Deque<Double> que){
		
		/*if(que.getLast()-que.getFirst()>0)
			return true;
		else
			return false;*/
		
		Deque<Double> pom = new ArrayDeque<Double>(que);
		boolean rastie[] = new boolean[pom.size()];
		double curValue = pom.pollFirst().doubleValue();
		
		for(int i=0;i<que.size()-1;i++){
			rastie [i]= curValue<pom.peekFirst().doubleValue();
			curValue = pom.pollFirst().doubleValue();
			
		}
		
		int h=0;
		for(int i=0;i<rastie.length;i++){
			if(rastie[i])
				h++;
		}
		
		if(h>=rastie.length/2)
			return true;
		else
			return false;
	}
	/**
	 *  spocita jednotlive prvky na vypocet smeru za poslednych que.size() tikov
	 * @param que rad integerov
	 * @return
	 */
	private int getTurnDirection(Deque<Integer> que){
		Iterator<Integer> i = que.iterator();
		int ret=0;
		while(i.hasNext())
			ret+= i.next();
		if(ret>0)
			return 1;
		else if(ret <0)
			return -1;
		else
			return 0;
	}
	
	
	
	/**
	 * Metoda, ktora vypocita uhol medzi aktualnou poziciou auta a bodom na trati pred autom.
	 * @param sm senzory
	 * @param positionDistance vzdialenost sledovaneho bodu od startu
	 * @param position vertikalna pozicia sledovaneho bodu na trati
	 * @return uhol
	 * @author Juraj Kosmel
	 */
	private double computeAngle(SensorModel sm, double positionDistance, double position) { 
		double alfa=0;
		double vertical=0;
		double horizontal=0;
		if((sm.getTrackPosition()>0 && position>0) || (sm.getTrackPosition()<0 && position<0)) { //nasa pozicia aj pozicia kde smerujeme je na rovnakej strane trate
			vertical=Math.abs((trackWidth/2.0)*Math.abs(sm.getTrackPosition())-trackWidth/2.0*Math.abs(position));
			horizontal=positionDistance-sm.getDistanceFromStartLine();
			alfa = Math.atan(vertical/horizontal);
		}
		else {
			vertical=Math.abs((trackWidth/2.0)*Math.abs(sm.getTrackPosition())+trackWidth/2.0*Math.abs(position)); //nasa pozicia a pozicia kde smerujeme su na roznych stranach trate
			horizontal=positionDistance-sm.getDistanceFromStartLine();
			alfa = Math.atan(vertical/horizontal);
		}
		return alfa;
	}
	
	/**
	 * Metoda na upravu smeru jazdy a navigaciu na konkretnu poziciu na trati.
	 * @param action aktualne nastavenie efektorov
	 * @param sm senzory
	 * @param positionDistance pozicia, na ktoru sa chceme dostat - kilometer od zaciatku trate
	 * @param position poloha na trati na danej pozicii
	 * @return upravena action - natocenie kolies
	 * @author Juraj Kosmel
	 */
	private Action navigateToPosition(Action action, SensorModel sm, double positionDistance, double position) {
		double alfa, beta, sum, diff;                                 //uhol natocenia k cielovej pozicii a uhol auta s osou, sucet a rozdiel uhlov
		alfa=computeAngle(sm, positionDistance, position); //vypocet uhla k miestu kde chceme smerovat
		beta = sm.getAngleToTrackAxis();				   //uhol nasho aktualneho smeru
		sum=Math.abs(alfa)+Math.abs(beta);				   //sucet uhlov
		diff=Math.abs(Math.abs(alfa)-Math.abs(beta));				   //rozdiel uhlov
		//System.out.println("beta: "+Math.toDegrees(beta)+" alfa: "+Math.toDegrees(alfa));
		
		if(beta>0) { //som otoceny vpravo
			if(sm.getTrackPosition() >= position) { //chcem ist vpravo vzhladom na moju poziciu
				if(Math.abs(beta)<=alfa) {
					action.steering=-(sum/0.36651943);
					//System.out.println("natoceny vpravo chcem vpravo beta mensia");
				}
				else {
					action.steering=diff/0.36651943;
					//System.out.println("natoceny vpravo chcem vpravo beta vacsia");
				}
			}
			else {								   //chcem ist vlavo vzhladom na moju poziciu
				action.steering=sum/0.36651943;
				//System.out.println("natoceny vpravo chcem vlavo");
			}
		}
		else {		 //som otoceny vlavo
			if(sm.getTrackPosition() >= position) { //chcem ist vpravo vzhladom na moju poziciu
				action.steering=-(sum/0.36651943);
				//System.out.println("natoceny vlavo chcem vpravo");
			}
			else {								   //chcem ist vlavo vzhladom na moju poziciu
				if(Math.abs(beta)<=alfa) {
					action.steering=sum/0.36651943;
					//System.out.println("natoceny vlavo chcem vlavo beta mensia");
				}
				else {
					action.steering=-(diff/0.36651943);
				//	System.out.println("natoceny vlavo chcem vlavo beta vacsia");
				}
			}
		}
		//System.out.println("steering: "+action.steering);
		return action;
	}
	
	
	
	
	
	/**
	 * Metoda urcujuca vztah medzi najdlhsim uhlom a natocenim volantu.
	 * @param longestAngle uhol k najvzdialenejsiemu okraju trate
	 * @return natocenie volantu
	 */
	private double steeringAdjustment(double longestAngle)
	{
		if (Math.abs(longestAngle) > 21)
			if (longestAngle > 0)
				return -1;
			else
				return 1;
		
		return -longestAngle / 21;
	}

	/**
	 * Metoda prebrana zo SimpleDriver sluziaca na navrat na trat.
	 * @param sm senzory
	 * @return akcia smerujuca k nabraniu vhodneho smeru
	 */
	private Action recoverToTrack(SensorModel sm)
	{
		//System.out.println("Prebieha zotavovanie z uviaznutia.");
		Action action = new Action();
		action.accelerate = 1;
					
		// zistime, ci je auto natocene spravnym smerom alebo smeruje von z trate
		if (sm.getAngleToTrackAxis() * sm.getTrackPosition() > 0)
		{
			action.gear = 1;
			action.steering = sm.getAngleToTrackAxis() / (Math.PI / 4);
		}
		else
		{
			action.gear = -1;
			action.steering = -sm.getAngleToTrackAxis() / (Math.PI / 4);
		}
		
		return action;
	}

	/**
	 * Metoda vrati priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 * v rozsahu <-90; 90>. Priemerovana je len jedna suvisla postupnost senzorov.
	 * @param trackEdgeSensors senzory okrajov trate
	 * @return priemer uhlov senzorov zaznamenavajucich najvzdialenejsi okraj trate
	 */
	private double computeAvgLongestSensorAngle(double[] trackEdgeSensors)
	{
		int max = 0, numberOfLongestSensors = 1;
		double angle = 0;

		// najde index prveho senzora v smere k najvzdialenejsiemu okraju trate
		for (int i = 1; i < trackEdgeSensors.length; i++)
			if (trackEdgeSensors[max] < trackEdgeSensors[i])
				max = i;
		
		// zisti kolko senzorov ma najvyssiu hodnotu
		while (max + numberOfLongestSensors < trackEdgeSensors.length &&
				trackEdgeSensors[max] == trackEdgeSensors[max + numberOfLongestSensors])
			numberOfLongestSensors++;

		// vypocet priemeru, vysledok je z rozsahu <-90; 90>
		for (int i = 0; i < numberOfLongestSensors; i++)
			angle = max + i;
		angle = (angle / numberOfLongestSensors) * 10 - 90;
		
		return angle;
	}

	/**
	 * Metoda vracajuca vhodnu hodnotu prevodoveho stupna.
	 * 
	 * Implementacia nie je urcena na zaradenie do "produkcneho" kodu.
	 * Jej cielom je uplna oddelenost od ostatnych prvkov ovladaca pre
	 * prehladnost vyvoja.
	 */
	private int setGear(SensorModel sensors)
	{
		int gear = sensors.getGear();

		int[]  gearUp = { 9500, 9400, 9500, 9500, 9500, 0 };
		int[]  gearDown = { 0, 3300, 6200, 7000, 7300, 7700 };
		
		double rpm = sensors.getRPM();
		if (gear < 1)
			gear = 1;

		// check if the RPM value of car is greater than the one suggested
		// to shift up the gear from the current one
		else if (gear < 6 && rpm >= gearUp[gear-1])
			gear++;
		
		// check if the RPM value of car is lower than the one suggested
		// to shift down the gear from the current one
		else if (gear > 1 && rpm <= gearDown[gear-1])
			gear--;
		
		return gear;
	}
	private double radiusFromDistance(double distance){
		if(distance>142)
			return 200;
		else if(distance>44)
			return 0.8107+1.0316*distance-0.0002*Math.pow(distance, 2);
		else if(distance>15)
			return 0.0579+1.0117*distance;
		else
			return 10.0168-0.6658*distance+0.0696*Math.pow(distance, 2);
	}
	private double steeringAdjustmentNew(double distance,double trackWidth,double strana){
		//TODO prve dva ifi na zaklade sirky trate mozno lepsie na zaklde pozicie na trati
		
		distance=distance-(trackWidth-6);
		
		if(distance<42)
			distance= distance*1.05;
		if(distance<15)
			distance= distance*1.1052;
		//System.out.println(distance);
		
		if(distance>138)
			return 0.0075;
		if(distance>43){
			strana= (0.8409/(distance-26.9920))*(1+(strana));
			//strana =0;
		}else if(distance>15){
			strana= (2.2216/(distance-9.9210))*(0.975+(strana));
		}else if (distance >6.3545)
			strana= (3.8193/(distance-6.3542))*(1+(strana));
		else
			strana =1; 
		if(strana>1)
			strana=1;
		if(strana<-1)
			strana=-1;
		
			
		return Math.abs(strana);
	}
	/*vypocita rychlost z steeringu*/
	private double countSpeed(double steering){
		if(steering>0.4){
			return 116.28-114.9683*steering+51.2030*Math.pow(steering, 2);
		}else if(steering>0.14){
			return 189.6921-506.5*steering+582.0934*Math.pow(steering, 2);
		}else if(steering>=0.1){
			return 268.5877-1508.3*steering+3798.4*Math.pow(steering, 2);
		}else if(steering>=0.018){
			return 22.4685/(steering+0.0471);
		}else
			return 500;
			
	}
	public void reset()
	{
		System.out.println("restarting the race");
	}

	public void shutdown()
	{
		
		System.out.println("shutdown");		
	}
}
