package sk.stuba.fiit.tp2010.team20.defence;

import java.util.HashMap;

import controller.model.TickInfo;
import champ2010client.Action;
import champ2010client.SensorModel;

/**
 * @author Vojtech Juhasz
 *ORP = Opponent's Relative Position
 */
enum ORP {LEFT, RIGHT, BEFORE, BEHIND, CRASH};

/**
 * @author Vojtech Juhasz
 */
public class OpponentDetector {
	static final double HDIST_TH = 2.0;
	static final double VDIST_TH = 1.0;
	static final double ANGLE_TRASHOLD = 0.5;
	static final double TURN_TRASHOLD = 0.5;
	static final double TRACK_EDGE = 0.95;
	static final int OVERTAKE_DISTANCE=20;
	static final int BLOCK_DISTANCE=100;
	static final int SAFE_DISTANCE = 5;
	
	static final int OPPONENT_SENSOR_COUNT = 36;
	private static HashMap<String, Opponent> opponentMap;
	private static double[] sensorAngles=null;
	
	static{
		opponentMap = new HashMap<String, Opponent>();
		
		sensorAngles = new double[OPPONENT_SENSOR_COUNT];
		double angle= -1*Math.PI;
		double angleDiff = 2*Math.PI/OPPONENT_SENSOR_COUNT;
		for(int i=0; i<sensorAngles.length; i++){
			sensorAngles[i] = angle;
			angle += angleDiff;
		}
	}

	public static Action monitor(TickInfo input){
		return monitor(input.getAction(), input.getSensorModel());
	}

	public static Action monitor(Action a, SensorModel sm){
		double[] oppDist = sm.getOpponentSensors();
		double trackPosition = sm.getTrackPosition();
		int closestOpponentIdx = -1;
		
		//look for opponents and determine the closest (if any)
		for(int i=0; i<oppDist.length; i++){
			if(oppDist[i]>=200)
				continue;
			if(closestOpponentIdx==-1)
				closestOpponentIdx=i;
			else
				if(oppDist[i]<oppDist[closestOpponentIdx])
					closestOpponentIdx=i;

			if(opponentMap.containsKey(""+i)){
				opponentMap.get(""+i).addPosition(new OpponentPosition(i, oppDist[i]));
			}else{
				opponentMap.put(""+i, new Opponent(i, oppDist[i]));
			}
//			System.out.print(opponentMap.get(""+i)+";");
//			System.out.print(opponentMap.get(""+i).getHorizontalDistance()+";");
		}
//		if(closestOpponentIdx!=-1)System.out.print("\n");
		
		if(closestOpponentIdx!=-1){
			if((trackPosition >= -1) || (trackPosition <= 1)){
				//the car is on the track
				Opponent nearestOpp = opponentMap.get(""+closestOpponentIdx);
				if(nearestOpp.getActualVerticalDistance()<0){
					//BLOCK
					if(nearestOpp.getRelativeSpeed()>=0 && 
							(-1*nearestOpp.getActualVerticalDistance())<BLOCK_DISTANCE){	
						//oponent sa priblizuje
						return blockOpponent(a, sm, nearestOpp);
					}else{
						//zanechavame oponenta
						return a;
					}
				}else if(nearestOpp.getActualVerticalDistance()>=0){
					//OVERTAKE
					if(nearestOpp.getRelativeSpeed()<=0 && 
							nearestOpp.getActualVerticalDistance()<OVERTAKE_DISTANCE){	
						//priblizujeme sa k oponentovi
						return overtakeOpponent(a, sm, nearestOpp);
					}else{
						//oponent sa vzdialuje
						return a;
					}
				}
			}else{
				//the car is recovering
				//TODO implement recovery defence
			}
		}else{
			//reset defence
			if(opponentMap.size()>0){
				opponentMap.clear();
				System.out.println("RESET-DEFENCE");
			}
		}
		return a;
	}

	private static Action blockOpponent(Action a, SensorModel sm, Opponent o){
		// TODO BLOCK-OPPONENT
		Action action = a;	
//		System.out.print("BLOCK: ");
//		double sd = a.steering;
//		double ac = a.accelerate;
		double sp, hd, ad, ot;
		double vd = -1*o.getActualVerticalDistance();
//		double angle = sm.getAngleToTrackAxis();

		if(Math.abs(vd)<=BLOCK_DISTANCE){
			ORP oppPos = o.getPositionBySensors();
			if(oppPos == ORP.BEHIND){
				action = a;
			}else if(oppPos==ORP.RIGHT){	// oponent je na pravej strane vozidla vodica
				if(isOnTrackEdge(sm)!=1){	//vodic NIE je na PRAVOM okraji trate
					ad = o.getActualDistance();
					hd = o.getActualHorizontalDistance();
					sp = o.getRelativeSpeed();
					
					ot = o.determineContactTime(sp, vd+SAFE_DISTANCE);
				}
			}else if(oppPos==ORP.LEFT){
				if(isOnTrackEdge(sm)!=-1){	//vodic NIE je na lavom kraji trate
//					sp = o.getBlockSpeed();
//					if(sp>40){
//						System.out.println(sp+" l2");
//					}else if(sp>15){
//						System.out.println(sp+" l1");
//					}else
//						System.out.println(sp+" l0");
				}
			}
		}
		
		return action;

//		System.out.print(vd+" | "+hd+" | "+a.steering+" | ");
//		if(vd<=BLOCK_DISTANCE && (sm.getTrackPosition()<1 && sm.getTrackPosition()>-1)){
//			sd += hd>0?-0.2:0.2;
//			ac += 0.1;
//			if(vd<=BLOCK_DISTANCE_L2 && (sm.getTrackPosition()<1 && sm.getTrackPosition()>-1)){
//				sd += hd>0?-0.1:0.1;
//				ac += 0.2;
//				if(vd<=BLOCK_DISTANCE_L3 && (sm.getTrackPosition()<1 && sm.getTrackPosition()>-1)){
//					if(hd>5){
//						sd += hd>0?-0.10:0.10;
//					}else if(hd>3){
//						sd += hd>0?-0.30:0.30;
//					}else{
//						sd += hd>0?-0.40:0.40;
//					}
//					ac += 0.3;
//					if(vd<=BLOCK_DISTANCE_L4 && (sm.getTrackPosition()<1 && sm.getTrackPosition()>-1)){
//						sd += hd>0?-0.10:0.10;
//						ac += 0.4;
//					}
//				}
//			}
//		}
//		a.accelerate = ac;
//		if(Math.abs(sd)>=TURN_TRASHOLD){
//			System.out.print("T\t");
//			a.steering = (sd/Math.abs(sd)) *TURN_TRASHOLD;
//		}else{
//			a.steering = sd;
//		}
//		System.out.print(a.steering+"\t"+a.accelerate+"\n");
//		return a;
		
	}

	static double radianToDegree(double radian){
		return radian*180/Math.PI;
	}
	static double degreeToRadian(double degree){
		return degree*Math.PI/180;
	}
	/**
	 * 
	 * @param sensor
	 * @return pociatocny uhol daneho senzora v radianoch v intervale &lt;-PI, PI&gt, alebo ak sensor&lt;0 alebo sensor&gt;<i>index posledneho senzora</i>, potom 10
	 */
	static double getSensorAngle(int sensor){
		if(sensor>=0 && sensor<sensorAngles.length)
			return sensorAngles[sensor];
		return 10;
	}

	/**
	 * 
	 * @param sm
	 * @return 0 ak autopilot nie je na kraji trate, 
	 * +1 ak autopilot sa nachadza na pravej strane alebo 
	 * -1 ak autpilot sa nachadza na lavej strane trate. 
	 */
	static int isOnTrackEdge(SensorModel sm){
		double p = Math.abs(sm.getTrackPosition());
		if(p>=OpponentDetector.TRACK_EDGE)
			return 1;
		else if(p<=-1*OpponentDetector.TRACK_EDGE)
			return -1;
		return 0;
	}
	/**
	 * TODO isCarDangerouslyTurned
	 * Tato funkcia urci ci je zatocenie auta nebezpecny. Vypocita to z
	 * aktualneho natocenia auta ku trate, aktualnej rychlosti a 
	 * @param sm
	 * @return
	 */
	static boolean isCarDangerouslyTurned(SensorModel sm){
		return false;
	}
	private static Action overtakeOpponent(Action a, SensorModel sm, Opponent o) {
		// TODO OVERTAKE-OPPONENT
//		System.out.println("OVERTAKE");
//		o.getBlockSpeed();
		return a;
	}

}
